/*******************************************************************************
 * Copyright (c) 2003 Berthold Daum.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors:
 *     Berthold Daum
 *******************************************************************************/
package com.bdaum.SpellChecker;

import java.io.IOException;
import java.net.URL;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.ui.plugin.AbstractUIPlugin;

import com.bdaum.SpellChecker.actions.CheckSpellingActionDelegate;
import com.bdaum.SpellChecker.preferences.SpellCheckerPreferences;

/**
 * This class cares for the initialization of preferences and
 * acts as a central registry for core components such as
 * SpellCheckManager and CheckSpellingActionDelegate.
 */
public class SpellCheckerPlugin extends AbstractUIPlugin {

  // Default dictionaries
  private static final String DEFAULTDICT = "dict/english.0";
  private static final String USERDICT = "dict/user.dict";

  // The singleton
  private static SpellCheckerPlugin plugin;
  // Default preferences
  private SpellCheckerPreferences preferences;
  // Active SpellCheckManager
  private SpellCheckManager manager;
  // Active ActionDelegate
  private CheckSpellingActionDelegate spellCheckingActionDelegate;
  // Spell checking thread
  private Thread spThread;
/**
   * The Constructor.
   */
  public SpellCheckerPlugin() {
    super();
    plugin = this;
    // Set configuration for jazzy engine. We make jazzy fetch
    // the configuration from our own configuration
    // implementation.
    System.setProperty("jazzy.config",
                    "com.bdaum.SpellChecker.SpellCheckConfiguration");
  }

  /**
   * Returns the plug-in singleton
   */
  public static SpellCheckerPlugin getDefault() {
    return plugin;
  }

  /**
   * Fetches the plug-in identification.
   * 
   * @return String  the identification.
   */
  public static String getId() {
    return getDefault().getBundle().getSymbolicName();
  }
  /**
   * Returns the absolute
   * path of the default dictionary file.
   * 
   * @return String - Default dictionary file path
   */
  public static String getDefaultDictionaryFileName() {
    return getDefaultDictionaryFileName(DEFAULTDICT);
  }

  /**
   * Returns the absolute
   * path of the default user dictionary file.
   * 
   * @return String - Default dictionary file path
   */
  public static String getDefaultUserDictFileName() {
    return getDefaultDictionaryFileName(USERDICT);
  }

  private static String getDefaultDictionaryFileName(
    String filename) {
    // First, fetch the URL of the plug-in.
    URL pluginURL = getDefault().getBundle().getEntry(filename);
    // This URL starts with the pseudo protocol "plugin:"
    // Therefore resolve this URL into a real URL
    try {
      URL resolvedURL = Platform.resolve(pluginURL);
      // Extract the path information
      return resolvedURL.getPath();
    } catch (IOException e) {
            logError(4,                    Messages.getString(
                        "SpellCheckerPlugin.Error_resolving_dictionary_URL"),
e);
    }
    return null;
  }
/**
   * Returns the preferences of this plug-in
   */
  public SpellCheckerPreferences getPreferences() {
    if (preferences == null)
      preferences = new SpellCheckerPreferences();
    return plugin.preferences;
  }

  /**
   * Initialization of the PreferenceStore
   * 
   * @param store - the plug-in's preference store
   */
  protected void initializeDefaultPreferences(
                                    IPreferenceStore store) {
      getPreferences().initializeDefaults(store);
  }
  /**
   * Returns the manager.
   * 
   * @return SpellCheckManager
   */
  public static SpellCheckManager getManager() {
    // Create SpellCheckManager instance if necessary
    if (plugin.manager == null)
      plugin.manager = new SpellCheckManager();
    return plugin.manager;
  }
  /**
   * Registers the active SpellCheckingActionDelegate.
   * 
   * @param delegate - active SpellCheckingActionDelegate
   */
  public static void setSpellCheckingActionDelegate(
                CheckSpellingActionDelegate delegate) {
    plugin.spellCheckingActionDelegate = delegate;
  }

  /**
   * Returns the currently active SpellCheckingActionDelegate.
   * 
   * @return CheckSpellingActionDelegate - active action
   *         delegate
   */
  public static CheckSpellingActionDelegate 
                                                        getSpellCheckingActionDelegate() {
    return plugin.spellCheckingActionDelegate;
  }
  /**
   * Checks if there is a pending spell checking thread
   * 
   * @return - true if a thread is pending
   */
  public static boolean isPending() {
    return (plugin.spThread != null && plugin.spThread.isAlive());
  }

  /**
   * Starts a new synchronous spell checking thread
   * 
   * @param thread - the new thread
   */
  public static void startThread(Thread thread) {
    plugin.spThread = thread;
    thread.start();
  }
   /**
   * Writes internal errors to log file.
   * 
   * @param code - Plug-in specific error code
   * @param message - message text
   * @param ex - Throwable that caused the error or null.
   */
  public static void logError(
    int code, String message, Throwable ex) {
    getDefault().getLog().log(
            new Status(IStatus.ERROR, getId(), code, message, ex));
  }

}
